require ("ioctl");
require ("select");
require ("termios");

private variable FIONREAD = 0x0000541B, TIOCMGET = 0x00005415, TIOCMBIS = 0x00005416,
  TIOCMBIC = 0x00005417, TIOCM_RTS = 0x004, TIOCM_DTR = 0x002, TIOCM_CTS = 0x020,
  TIOCM_CD = 0x040, TIOCM_RI = 0x080, TIOCM_DSR = 0x100;

typedef struct
{
   baudrate, bytesize, fd, input_baudrate, output_baudrate, parity, port,
     rtscts, stopbits, timeout, xonxoff, write_timeout
} Serial_Type;

define serial_open (port)
{
   variable tty, cflag, iflag, lflag, oflag, cc, s = @Serial_Type;
   
   s.baudrate = qualifier ("baudrate", 9600);
   s.bytesize = qualifier ("bytesize", 8);
   s.input_baudrate = qualifier ("input_baudrate", 0);
   s.output_baudrate = qualifier ("output_baudrate", 0);
   s.parity = qualifier ("parity", "none");
   s.port = port;
   s.rtscts = qualifier ("rtscts", 0);
   s.stopbits = qualifier ("stopbits", 1);
   s.timeout = qualifier ("timeout", 1);
   s.xonxoff = qualifier ("xonxoff", 0);
   s.write_timeout = qualifier ("write_timeout", 1);
   
   s.fd = open (port, O_RDWR | O_NOCTTY | O_NONBLOCK);
   
   tty = tcgetattr (s.fd);

   if (s.input_baudrate == 0)
     () = cfsetispeed (tty, s.baudrate);
   else
     () = cfsetispeed (tty, s.input_baudrate);

   if (s.output_baudrate == 0)
     () = cfsetospeed (tty, s.baudrate);
   else
     () = cfsetospeed (tty, s.output_baudrate);      

   cflag = termios_get_cflag (tty);
   cflag |= CREAD | CLOCAL;
   cflag &= ~CSIZE;
   switch (s.bytesize)
     { case 5 : cflag |= CS5;}
     { case 6 : cflag |= CS6;}
     { case 7 : cflag |= CS7;}
     { case 8 : cflag |= CS8;}
     { throw InvalidParmError;}

   switch (s.stopbits)
     { case 1 : cflag &= ~CSTOPB;}
     { case 2 : cflag |= CSTOPB;}
     {throw InvalidParmError;}

   switch (s.parity)
     { case "none" : cflag &= ~(PARENB | PARODD);}
     { case "odd" : cflag &= ~PARODD; cflag |= PARENB;}
     { case "even" : cflag |= (PARENB | PARODD);}
     { throw InvalidParmError;}

   if (s.rtscts)
     cflag |= CRTSCTS;
   else
     cflag &= ~CRTSCTS;
   
   iflag = termios_get_iflag (tty);   
   iflag &= ~(INLCR | IGNCR | ICRNL | IGNBRK | PARMRK | INPCK | ISTRIP);

#ifdef IXANY
   if (s.xonxoff)
     iflag |= (IXON | IXOFF | XANY);   
   else
     iflag &= ~(IXON | IXOFF | XANY);
#else
   if (s.xonxoff)
     iflag |= (IXON | IXOFF);   
   else
     iflag &= ~(IXON | IXOFF);     
#endif
     
   lflag = termios_get_lflag (tty);
   lflag &= ~(ICANON | ECHO | ECHOE | ECHOK | ECHONL | ISIG | IEXTEN);
   
   oflag = termios_get_oflag (tty);
   oflag &= ~OPOST;
   
   termios_set_cflag (tty, cflag);
   termios_set_iflag (tty, iflag);
   termios_set_lflag (tty, lflag);   
   termios_set_oflag (tty, oflag);
      
   cc = termios_get_cc (tty);
   cc [VMIN] = 0;
   cc [VTIME] = 0;
   termios_set_cc (tty, cc);
   
   () = tcsetattr (s.fd, TCSANOW, tty);   
   () = tcflush (s.fd, TCIOFLUSH);   

   return s;
}

define serial_close (s)
{
   ifnot (FD_Type == typeof (s.fd))
     throw OpenError;
   () = tcflush (s.fd, TCIOFLUSH);
   if (-1 == close (s.fd))
     throw IOError;
}

define serial_read ()
{
   variable buf, fds, data = "", n = 1, s;
   
   if (_NARGS == 2)
     {	
	n = ();
	s = ();
     }   
   else if (_NARGS == 1)
     s = ();
   else
     throw NumArgsError;
   
   ifnot (FD_Type == typeof (s.fd))
     throw OpenError;
   
   while (bstrlen (data) < n)
     {
	fds = select ([s.fd], [NULL], [NULL], s.timeout);
	ifnot (fds.nready)
	  break;
	() = read (s.fd, &buf, n - bstrlen (data));
	data += buf;
	if (s.timeout && (buf != ""))
	  break;
     }
   return data;
}

define serial_write (s, data)
{
   variable buf, fds, len, n;
   
   ifnot (FD_Type == typeof (s.fd))
     throw OpenError;
   buf = data [*];
   len = bstrlen (data);
   while (len > 0)
     {
	if (s.write_timeout)
	  {	     
	     fds = select ([NULL], [s.fd], [NULL], s.write_timeout);
	     ifnot (fds.nready)
	       break;
	  }	
	n = write (s.fd, buf);
	if (s.write_timeout)
	  {	     
	     fds = select ([NULL], [s.fd], [NULL], s.write_timeout);
	     ifnot (fds.nready)
	       break;
	  }	
	buf = buf [[n:]];
	len = len - n;
     }   
}

define flush_input (s)
{
   ifnot (FD_Type == typeof (s.fd))
     throw OpenError;
   () = tcflush (s.fd, TCIFLUSH);
}

define flush_output (s)
{
   ifnot (FD_Type == typeof (s.fd))
     throw OpenError;
   () = tcflush (s.fd, TCOFLUSH);
}

define serial_in_waiting (s)
{
   variable n;
    
   (, n) = ioctl (s.fd, FIONREAD, pack ("i", 0));
   return unpack ("i", n);
}

define serial_send_break (s, t)
{
   tcsendbreak (s, t);
}

define serial_set_rts (s, l)
{
   if (l)
     () = ioctl (s.fd, TIOCMBIS, pack ("i", TIOCM_RTS));
   else
     () = ioctl (s.fd, TIOCMBIC, pack ("i", TIOCM_RTS));
}

define serial_set_dtr (s, l)
{
   if (l)
     () = ioctl (s.fd, TIOCMBIS, pack ("i", TIOCM_DTR));
   else
     () = ioctl (s.fd, TIOCMBIC, pack ("i", TIOCM_DTR));   
}

define serial_get_cts (s)
{
   variable n;
    
   (, n) = ioctl (s.fd, TIOCMGET, pack ("i", 0));
   return (unpack ("i", n) & TIOCM_CTS) != 0;   
}

define serial_get_dsr (s)
{
   variable n;
    
   (, n) = ioctl (s.fd, TIOCMGET, pack ("i", 0));
   return (unpack ("i", n) & TIOCM_DSR) != 0;   
}

define serial_get_ri (s)
{
   variable n;
    
   (, n) = ioctl (s.fd, TIOCMGET, pack ("i", 0));
   return (unpack ("i", n) & TIOCM_RI) != 0;   
}

define serial_get_cd (s)
{
   variable n;
    
   (, n) = ioctl (s.fd, TIOCMGET, pack ("i", 0));
   return (unpack ("i", n) & TIOCM_CD) != 0;
}

require ("serial_util");
